/**************************************************************************
 *
 * ashby2014.c
 * Source code for solving the model described in the main text of 
 * "Parasitic castration promotes coevolutionary cycling but also imposes a
 * cost on sex" by Ashby and Gupta. This solver uses a Runge-Kutta 
 * algorithm. The source code is written in C#, but is implemented using 
 * the MEX functionality in MATLAB (R2012a). Compile using 
 * "mex ashby2014.c" at the MATLAB command line and call using either:
 * 
 * [t,x,EQFLAG] = ashby2014
 *    (MaxTime,r,f,beta,mu,alpha,rho,m,init_pop,exttol,eqtol,solvertol,0);
 * 
 * for reduced output, or:
 * 
 * [t,x,EQFLAG] = ashby2014
 *    (MaxTime,r,f,beta,mu,alpha,rho,m,init_pop,exttol,eqtol,solvertol,1);
 * 
 * for the full time series data. 
 *
 * %%% Inputs: %%%
 *
 * MaxTime      number of time units
 * r            maximum per-capita growth rate for hosts
 * f            fertility of infected hosts (between 0 and 1, inclusive)
 * beta         transmission rate
 * mu           natural mortality rate
 * alpha        disease-associated mortality rate
 * rho          recombination rate
 * m            number of mating attempts for sexuals
 * init_pop     row vector containing initial population sizes (100 entries)
 * exttol       exit simulation if the sexual population falls below this threshold
 * eqtol        allows the solver to exit early if dynamics have nearly reached equilibrium
 * solvertol    Adjusts the accuracy of the solver (smaller=more accurate, but takes longer)
 * output       0 = only output final population state
 *              1 = output full time series data
 *
 * %%% Outputs: %%%
 * 
 * t            time vector
 * x            time series data. Rows correspond to the time points in "t"
 *              Colums arranged in blocks of ten by host genotype:
 *              
 *              00/00,01/01,10/10,11/11,00/01,00/10,00/11,01/10,01/11,10/11
 *
 *              These blocks are arranged as follows:
 *              1) Susceptible (sexual)
 *              2) Susceptible (asexual)
 *              3) Infected by parasite 00 (sexual)
 *              4) Infected by parasite 01 (sexual)
 *              5) Infected by parasite 10 (sexual)
 *              6) Infected by parasite 11 (sexual)
 *              7) Infected by parasite 00 (asexual)
 *              8) Infected by parasite 01 (asexual)
 *              9) Infected by parasite 10 (asexual)
 *              10) Infected by parasite 11 (asexual)
 *
 * EQFLAG       0 = hasn't reached equilibrium
 *              1 = equilibrium reached
 *
 *  Ben Ashby 
 *  31/03/2014
 *************************************************************************/

#include "mex.h"
#include "math.h"
struct PARAM{
    double r; /* Per-capita birth rate */
    double f; /* Fertility of infected hosts */
    double beta; /* Transmission rate */
    double mu; /* Natural mortality rate */
    double alpha; /* Disease-associated death rate */
    double rho; /* Recombination rate */
    double m; /* Number of mating attempts for sexuals */
    double exttol; /* Exit simulation if the sexual population falls below this threshold */
    double eqtol; /* Allows the solver to exit early if dynamics have nearly reached equilibrium */
    double solvertol; /* Adjusts the accuracy of the solver (smaller=more accurate, but takes longer) */
    int neq; /* Number of equations */
};

/* Function definitions */
void ddt(double *,double *,struct PARAM *);
void rkqs(double *,double *,double *,double *,double *,struct PARAM *);
void rkck(double *,double *,double *,double *,double,struct PARAM *);
double FMAX(double,double);
double FMIN(double,double);
        
#define TINY 1.0e-30
void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    double *init_pop, *x, *dxdt, upper, lower, *yscale, *parameter, *EQFLAG, *yArray, *tArray, *tArraytemp, *yArraytemp, h[1], hnext[1];
    double t, MaxTime, totalpop, intervallength, intervalend, minsex, maxsex, sex, diff, mean;
    int i, j, colLen, count, exitflag, flag, output;
    struct PARAM p;
    
    p.neq = 100; /* Total number of equations */
        
    x = (double *)malloc(p.neq*sizeof(double));
    dxdt = (double *)malloc(p.neq*sizeof(double));
    yscale = (double *)malloc(p.neq*sizeof(double));
    
    /* Use code below for terminating solution early (approximation to save time) */
    intervallength = 1000;
    intervalend = intervallength;
    if(nrhs!=13){
        printf("Incorrect number of input arguments, using defaults...\n");
        MaxTime = 100;
        p.r = 1;
        p.f = 0;
        p.beta = 0.05;
        p.mu = 0.1;
        p.alpha = 2;
        p.rho = 0;
        p.m = 10;
        
        /* Set up initial population */
        for (i=0;i<20;i++) x[i] = 1;
        for (i=20;i<p.neq;i++) x[i] = 0.01*i;

        p.exttol = 0;
        p.eqtol = 1e-4;
        p.solvertol = 1e-6;
        output = 1;
    }
    else{ /* Use input arguments as parameters */
        parameter= mxGetPr(prhs[0]);
        MaxTime= *parameter;
        parameter= mxGetPr(prhs[1]);
        p.r= *parameter;
        parameter= mxGetPr(prhs[2]);
        p.f= *parameter;
        parameter= mxGetPr(prhs[3]);
        p.beta= *parameter;
        parameter= mxGetPr(prhs[4]);
        p.mu= *parameter;
        parameter= mxGetPr(prhs[5]);
        p.alpha= *parameter;
        parameter= mxGetPr(prhs[6]);
        p.rho= *parameter;
        parameter= mxGetPr(prhs[7]);
        p.m= *parameter;
        init_pop= mxGetPr(prhs[8]);
        parameter= mxGetPr(prhs[9]);
        p.exttol= *parameter;
        parameter= mxGetPr(prhs[10]);
        p.eqtol= *parameter;
        parameter= mxGetPr(prhs[11]);
        p.solvertol = *parameter;
        parameter= mxGetPr(prhs[12]);
        output= (int)*parameter;
        
        for (i=0;i<p.neq;i++) x[i] = init_pop[i];        
    }
    
    /*upper and lower are used to determine equilibrium status */
    upper = 2*x[0]+x[4]+x[5]+x[6];
    lower = 2*x[0]+x[4]+x[5]+x[6];
    
    /* Outputs */
    if(output>0){ /* Return full output */
        colLen = (int)(FMIN(1e7, MaxTime*1000));
        tArraytemp = (double *)malloc(colLen * sizeof(double));
        yArraytemp = (double *)malloc(colLen*p.neq*sizeof(double));
        tArraytemp[0] = 0;
        for (i=0;i<p.neq;i++) yArraytemp[i*colLen] = x[i];
    }
    else{
        colLen = (int)1e8; /* Controls the maximum number of steps */
    }
    plhs[2] = mxCreateDoubleMatrix(1, 1, mxREAL);
    EQFLAG = mxGetPr(plhs[2]);
    EQFLAG[0]=0;
    
    
    /* Misc parameters */
    t=0;
    count=0;
    h[0] = 1e-3;
    hnext[0] = 1e-3;
    exitflag = 1;
    count=0;
    flag = 0;    
    minsex = 0.5;
    maxsex = 0.5;
    
    /* Solver */
    do{
        if(1.1*hnext[0]>(MaxTime-t)){
            hnext[0] = MaxTime-t;
            h[0] = MaxTime-t;
            t=MaxTime;
        }
        else{
            h[0] = hnext[0];
            t+=h[0];
        }
        if(t==MaxTime) exitflag=0;
        ddt(x, dxdt, &p);
        /*   Scaling used to monitor accuracy. This general-purpose choice can be modified if need be. */
        for (i=0;i<p.neq;i++) yscale[i]=fabs(x[i])+fabs(dxdt[i]*(*h))+TINY;
        rkqs(x, dxdt, h, hnext, yscale, &p);
        
        /*Make sure things are always non-negative, and update upper/lower */
        for(i=0;i<p.neq;i++) {
            x[i] = FMAX(x[i], 0);
        }
        upper = FMAX(upper,2*x[0]+x[4]+x[5]+x[6]);
        lower = FMIN(lower,2*x[0]+x[4]+x[5]+x[6]);
        totalpop = 0;
        for(i=0;i<p.neq;i++) {
            x[i] = FMAX(x[i], 0);
            totalpop+=x[i];
        }
        sex = 0;
        for(i=0;i<10;i++){
            sex+=x[i];
        }
        for(i=20;i<60;i++){
            sex+=x[i];
        }
        sex/=totalpop; /* Frequency of sex */
        minsex = FMIN(minsex,sex);
        maxsex = FMAX(maxsex,sex);
        if(sex<p.exttol){ /* Sex has gone extinct */
            EQFLAG[0]=1;
            break;
        }
        
        /* Track time series */
        count++;
        if(output>0){
            for (i=0;i<p.neq;i++) yArraytemp[count + i*colLen] = x[i];
            tArraytemp[count]= t;
        }
        
        /* Equilibrium checker */
        if(t>=intervalend){
           if(fabs((maxsex-minsex)/(minsex+(maxsex-minsex)/2))<p.eqtol) {
                EQFLAG[0]=1;
                break;
            }
           
            diff = fabs(upper-lower);
            mean = lower+diff/2;
            if(mean>1e-10 && (diff/mean)<p.eqtol){
                EQFLAG[0] = 1;
                break;
            }
            maxsex = sex;
            minsex = sex;
            upper = 2*x[0]+x[4]+x[5]+x[6];
            lower = 2*x[0]+x[4]+x[5]+x[6];
            intervalend=FMIN(intervalend+intervallength, MaxTime);
        }
    }while(count<(colLen-1) && t<=MaxTime && exitflag);
    
    /* Create remaining outputs */
    if(output>0){
        /* Create actual output arrays now that we know the number of steps taken */
        plhs[0] = mxCreateDoubleMatrix(count+1, 1, mxREAL);
        plhs[1] = mxCreateDoubleMatrix(count+1, p.neq, mxREAL);
        tArray = mxGetPr(plhs[0]);
        yArray = mxGetPr(plhs[1]);
        for (i=0;i<=count;i++){
            tArray[i] = tArraytemp[i];
            for (j=0;j<p.neq;j++) yArray[i + j*(count+1)] = yArraytemp[i + j*colLen];
        }
        free(tArraytemp);
        free(yArraytemp);
    }
    else{
        plhs[0] = mxCreateDoubleMatrix(1, 1, mxREAL);
        plhs[1] = mxCreateDoubleMatrix(1, p.neq, mxREAL);
        tArray = mxGetPr(plhs[0]);
        yArray = mxGetPr(plhs[1]);
        tArray[0] = t;
        for (j=0;j<p.neq;j++) yArray[j] = x[j];        
    }
    
    /* Free memory */
    free(x);
    free(dxdt);
    free(yscale);
    return;
}

void ddt(double *x,double *dev,struct PARAM *p)
{
    double F[4],FOI[4], temp[4], temp1[10], births_sex[10], births_asex[10], deaths_sex_sus[10], deaths_asex_sus[10], infections_sex_sus[10], infections_asex_sus[10], infections_sex_inf[40], infections_asex_inf[40], deaths_sex_inf[40], deaths_asex_inf[40],Q[40];
    double NS, z, SF, tempsum1, tempsum2;
    int i,j,count;
    
    /* Define specificity */
    for(i=0;i<40;i++) Q[i]=0;
    Q[0]=1; Q[11]=1; Q[22]=1; Q[33]=1;
    Q[4]=0.5; Q[5]=0.5; Q[6]=0.5; Q[14]=0.5; Q[17]=0.5; Q[18]=0.5; Q[25]=0.5;  Q[27]=0.5; Q[29]=0.5; Q[36]=0.5; Q[38]=0.5; Q[39]=0.5;
    
    NS = 0;
    for(i=0;i<10;i++){
        NS+=x[i];
    }
    for(i=20;i<60;i++){
        NS+=x[i];
    }
    tempsum1=0;
    if(NS==0){
        for(i=0;i<10;i++){
            births_sex[i]=0;
        }
    }
    else{    
        /* 'Healthy' haplotype frequencies */
        F[0] = 2*x[0]+x[4]+x[5]+x[6] + p->rho*(x[7]-x[6]);
        F[1] = 2*x[1]+x[4]+x[7]+x[8] + p->rho*(x[6]-x[7]);
        F[2] = 2*x[2]+x[5]+x[7]+x[9] + p->rho*(x[6]-x[7]);
        F[3] = 2*x[3]+x[6]+x[8]+x[9] + p->rho*(x[7]-x[6]);
        for(j=0;j<4;j++){ /* Recombination */
            F[0] += p->f*(2*x[0+j*10+20]+x[4+j*10+20]+x[5+j*10+20]+x[6+j*10+20] + p->rho*(x[7+j*10+20]-x[6+j*10+20]));
            F[1] += p->f*(2*x[1+j*10+20]+x[4+j*10+20]+x[7+j*10+20]+x[8+j*10+20] + p->rho*(x[6+j*10+20]-x[7+j*10+20]));
            F[2] += p->f*(2*x[2+j*10+20]+x[5+j*10+20]+x[7+j*10+20]+x[9+j*10+20] + p->rho*(x[6+j*10+20]-x[7+j*10+20]));
            F[3] += p->f*(2*x[3+j*10+20]+x[6+j*10+20]+x[8+j*10+20]+x[9+j*10+20] + p->rho*(x[7+j*10+20]-x[6+j*10+20]));
        }
        SF=F[0]+F[1]+F[2]+F[3];
        if(SF==0){
            z = 0;
        }
        else{ /* Castration and multiple mating */
            z=1-FMIN(1,pow(FMIN(1,FMAX(0,1-(SF/(2*NS)))),p->m));
            for(i=0;i<4;i++){
                births_sex[i] = (p->r*z/(2*SF))*F[i]*F[i];
            }
            /* NB: Not dividing by 2 because the rest are heterozygotes */
            births_sex[4] = (p->r*z/SF)*F[0]*F[1];
            births_sex[5] = (p->r*z/SF)*F[0]*F[2];
            births_sex[6] = (p->r*z/SF)*F[0]*F[3];
            births_sex[7] = (p->r*z/SF)*F[1]*F[2];
            births_sex[8] = (p->r*z/SF)*F[1]*F[3];
            births_sex[9] = (p->r*z/SF)*F[2]*F[3];
        }
    }
    
    /* Asexual births   */  
    tempsum1=0;
    tempsum2=0;
    for(i=0;i<10;i++){
        births_asex[i] = x[10+i];
        for(j=0;j<4;j++){
            births_asex[i]+=p->f*x[60+i+j*10];
        }
        births_asex[i] = p->r*births_asex[i];
        temp1[i] = births_asex[i];
    }
    
    /* Deaths */
    for(i=0;i<10;i++){
        deaths_sex_sus[i] = p->mu*x[i];
        deaths_asex_sus[i] = p->mu*x[10+i];
    }
    for(i=0;i<40;i++){
        deaths_sex_inf[i] = (p->mu+p->alpha)*x[20+i];
        deaths_asex_inf[i] = (p->mu+p->alpha)*x[60+i];
    }
    
    /* Force of infection & mutations/external FOI */
    for(j=0;j<4;j++){
        FOI[j]=0;
        for(i=0;i<10;i++){
            FOI[j]+=x[20+i+j*10]+x[60+i+j*10];
        }
        FOI[j]=p->beta*FOI[j];
        temp[j]=FOI[j];
    }

    /* Infections */
    count=0;
    for(i=0;i<10;i++){
        infections_sex_sus[i] = 0;
        infections_asex_sus[i] = 0;
    }
    for(j=0;j<4;j++){
        for(i=0;i<10;i++){
            infections_sex_inf[count]=Q[count]*FOI[j]*x[i];
            infections_asex_inf[count]=Q[count]*FOI[j]*x[10+i];
            infections_sex_sus[i]+=infections_sex_inf[count];
            infections_asex_sus[i]+=infections_asex_inf[count];   
            count++;
        }
    }
    
    /* Update equations */
    for(i=0;i<10;i++){
        dev[i] = births_sex[i] - deaths_sex_sus[i] - infections_sex_sus[i];
        dev[10+i] = births_asex[i] - deaths_asex_sus[i] - infections_asex_sus[i];
    }
    i=0;   
    for(i=0;i<40;i++){
        dev[20+i] = infections_sex_inf[i] - deaths_sex_inf[i];
        dev[60+i] = infections_asex_inf[i] - deaths_asex_inf[i];
    }
}

void rkqs(double *y,double *dydt,double *h,double *hnext,double *yscale,struct PARAM *p)
{
    double *ytemp, *yerr;
    double htemp,errmax;
    int i,n,attempts;
    
    n= p->neq;
    ytemp= (double *)malloc(n * sizeof(double));
    yerr=  (double *)malloc(n * sizeof(double));
    
    htemp= *h;
    
    rkck(y,dydt,ytemp,yerr,*h,p);
    *hnext= FMAX(*h,p->solvertol);    
    
    attempts=0;
    for(;;)
    {
        attempts++;
        rkck(y,dydt,ytemp,yerr,*h,p);
        errmax= 0.0;
        for(i=0;i<n;i++) {
            errmax= FMAX(errmax,fabs(yerr[i]/yscale[i]));
        }
        errmax/= p->solvertol;
        if(errmax<=1.0) break;
        htemp= 0.9*(*h)*pow(errmax,-0.25);
        *h= (*h>=0.0 ? FMAX(htemp,0.1*(*h)) : FMIN(htemp,0.1*(*h)));
        if(attempts>=10000){
            for(i=0;i<n;i++) printf("%f,",yerr[i]);
                
            printf("\n%f\n",errmax);
            mexErrMsgTxt("Error! Can't reduce step size further!\n");
            *hnext = -1;
            break;
        }
    }
    if(*hnext>0){
        if(errmax > 1.89E-4) {
            *hnext= 0.9*(*h)*pow(errmax, -0.2);
        }
        else {
            *hnext= 5.0*(*h);
        }
    }
    
    for(i=0;i<n;i++) y[i]= ytemp[i];
    
    free(ytemp);
    free(yerr);
}

void rkck(double *y,double *dydt,double *yout,double *yerr,double h,struct PARAM *p)
{
    int i,n=p->neq;
    double *k2,*k3,*k4,*k5,*k6,*ytemp;
    static double b21=0.2,b31=3.0/40.0,b32=9.0/40.0,b41=0.3,b42=-0.9,b43=1.2,
            b51=-11.0/54.0,b52=2.5,b53=-70.0/27.0,b54=35.0/27.0,b61=1631.0/55296,
            b62=175.0/512.0,b63=575.0/13824.0,b64=44275.0/110592,b65=253.0/4096.0,
            c1=37.0/378.0,c3=250.0/621.0,c4=125.0/594.0,c6=512.0/1771.0,dc5=-277.00/14336;
    double dc1=c1-2825.0/27648.0,dc3=c3-18575.0/48384.0,dc4=c4-13525.0/55296.0,
            dc6=c6-0.25;
    
    k2= (double *)malloc(n * sizeof(double));
    k3= (double *)malloc(n * sizeof(double));
    k4= (double *)malloc(n * sizeof(double));
    k5= (double *)malloc(n * sizeof(double));
    k6= (double *)malloc(n * sizeof(double));
    ytemp= (double *)malloc(n * sizeof(double));
    
    for(i=0;i<n;i++)
        ytemp[i]= y[i]+b21*h*dydt[i];
    ddt(ytemp,k2,p);
    
    for(i=0;i<n;i++)
        ytemp[i]= y[i]+h*(b31*dydt[i]+b32*k2[i]);
    ddt(ytemp,k3,p);
    
    for(i=0;i<n;i++)
        ytemp[i]= y[i]+h*(b41*dydt[i]+b42*k2[i]+b43*k3[i]);
    ddt(ytemp,k4,p);
    
    for(i=0;i<n;i++)
        ytemp[i]= y[i]+h*(b51*dydt[i]+b52*k2[i]+b53*k3[i]+b54*k4[i]);
    ddt(ytemp,k5,p);
    
    for(i=0;i<n;i++)
        ytemp[i]= y[i]+h*(b61*dydt[i]+b62*k2[i]+b63*k3[i]+b64*k4[i]+b65*k5[i]);
    ddt(ytemp,k6,p);
    
    for(i=0;i<n;i++)
        yout[i]= y[i]+h*(c1*dydt[i]+c3*k3[i]+c4*k4[i]+c6*k6[i]);
    for(i=0;i<n;i++)
        yerr[i]= h*(dc1*dydt[i]+dc3*k3[i]+dc4*k4[i]+dc5*k5[i]+dc6*k6[i]);
    
    free(k2);
    free(k3);
    free(k4);
    free(k5);
    free(k6);
    free(ytemp);
}

double FMAX(double l,double r)
{
    if(l>r)return l;
    else   return r;
}

double FMIN(double l,double r)
{
    if(l<r)return l;
    else   return r;
}